import pytest
import xarray as xr

import xdggs.utils


@pytest.mark.parametrize("name", ["h3", "s2", "rhealpix", "healpix"])
def test_register_dggs(monkeypatch, name):
    registry = {}

    monkeypatch.setattr(xdggs.utils, "GRID_REGISTRY", registry)

    grid = object()
    registered_grid = xdggs.utils.register_dggs(name)(grid)

    assert grid is registered_grid

    assert len(registry) == 1
    assert name in registry and registry[name] is grid


@pytest.mark.parametrize(
    ["variables", "expected"],
    (
        (
            {"cell_ids": xr.Variable("cells", [0, 1])},
            ("cell_ids", "cells"),
        ),
        (
            {"zone_ids": xr.Variable("zones", [0, 1])},
            ("zone_ids", "zones"),
        ),
    ),
)
def test_extract_cell_id_variable(variables, expected):
    expected_var = variables[expected[0]]

    actual = xdggs.utils._extract_cell_id_variable(variables)
    actual_var = actual[1]

    assert actual[::2] == expected
    xr.testing.assert_equal(actual_var, expected_var)



import pytest
import xarray as xr

import xdggs


@pytest.mark.parametrize(
    ["obj", "grid_info", "name"],
    (
        pytest.param(
            xr.Dataset(
                coords={
                    "cell_ids": (
                        "cells",
                        [1],
                        {
                            "grid_name": "healpix",
                            "level": 1,
                            "indexing_scheme": "ring",
                        },
                    )
                }
            ),
            None,
            None,
            id="dataset-from attrs-standard name",
        ),
        pytest.param(
            xr.DataArray(
                [0.1],
                coords={
                    "cell_ids": (
                        "cells",
                        [1],
                        {
                            "grid_name": "healpix",
                            "level": 1,
                            "indexing_scheme": "ring",
                        },
                    )
                },
                dims="cells",
            ),
            None,
            None,
            id="dataarray-from attrs-standard name",
        ),
        pytest.param(
            xr.Dataset(
                coords={
                    "zone_ids": (
                        "zones",
                        [1],
                        {
                            "grid_name": "healpix",
                            "level": 1,
                            "indexing_scheme": "ring",
                        },
                    )
                }
            ),
            None,
            "zone_ids",
            id="dataset-from attrs-custom name",
        ),
        pytest.param(
            xr.Dataset(coords={"cell_ids": ("cells", [1])}),
            {"grid_name": "healpix", "level": 1, "indexing_scheme": "ring"},
            None,
            id="dataset-dict-standard name",
        ),
    ),
)
def test_decode(obj, grid_info, name) -> None:
    kwargs = {}
    if name is not None:
        kwargs["name"] = name

    if isinstance(grid_info, dict):
        expected_grid_info = grid_info
    elif isinstance(grid_info, xdggs.DGGSInfo):
        expected_grid_info = grid_info.to_dict()
    else:
        expected_grid_info = obj[name if name is not None else "cell_ids"].attrs

    actual = obj.dggs.decode(grid_info, **kwargs)
    assert any(isinstance(index, xdggs.DGGSIndex) for index in actual.xindexes.values())
    assert actual.dggs.grid_info.to_dict() == expected_grid_info


@pytest.mark.parametrize(
    ["obj", "expected"],
    (
        (
            xr.DataArray(
                [0],
                coords={
                    "cell_ids": (
                        "cells",
                        [3],
                        {
                            "grid_name": "healpix",
                            "level": 1,
                            "indexing_scheme": "ring",
                        },
                    )
                },
                dims="cells",
            ),
            xr.Dataset(
                coords={
                    "latitude": ("cells", [66.44353569089877]),
                    "longitude": ("cells", [315.0]),
                }
            ),
        ),
        (
            xr.Dataset(
                coords={
                    "cell_ids": (
                        "cells",
                        [0x832830FFFFFFFFF],
                        {"grid_name": "h3", "level": 3},
                    )
                }
            ),
            xr.Dataset(
                coords={
                    "latitude": ("cells", [38.19320895]),
                    "longitude": ("cells", [-122.19619676]),
                }
            ),
        ),
    ),
)
def test_cell_centers(obj, expected):
    obj_ = obj.pipe(xdggs.decode)

    actual = obj_.dggs.cell_centers()

    xr.testing.assert_allclose(actual, expected)


@pytest.mark.parametrize(
    ["obj", "expected"],
    (
        (
            xr.DataArray(
                [0],
                coords={
                    "cell_ids": (
                        "cells",
                        [3],
                        {
                            "grid_name": "healpix",
                            "level": 1,
                            "indexing_scheme": "ring",
                        },
                    )
                },
                dims="cells",
            ),
            xr.DataArray(
                [0],
                coords={
                    "latitude": ("cells", [66.44353569089877]),
                    "longitude": ("cells", [315.0]),
                    "cell_ids": (
                        "cells",
                        [3],
                        {
                            "grid_name": "healpix",
                            "level": 1,
                            "indexing_scheme": "ring",
                        },
                    ),
                },
                dims="cells",
            ),
        ),
        (
            xr.Dataset(
                coords={
                    "cell_ids": (
                        "cells",
                        [0x832830FFFFFFFFF],
                        {"grid_name": "h3", "level": 3},
                    )
                }
            ),
            xr.Dataset(
                coords={
                    "latitude": ("cells", [38.19320895]),
                    "longitude": ("cells", [-122.19619676]),
                    "cell_ids": (
                        "cells",
                        [0x832830FFFFFFFFF],
                        {"grid_name": "h3", "level": 3},
                    ),
                }
            ),
        ),
    ),
)
def test_assign_latlon_coords(obj, expected):
    obj_ = obj.pipe(xdggs.decode)

    actual = obj_.dggs.assign_latlon_coords()

    xr.testing.assert_allclose(actual, expected)



import pytest

from xdggs import tutorial


@pytest.mark.parametrize(
    ["ds_name", "grid_name"],
    (
        ("air_temperature", "h3"),
        ("air_temperature", "healpix"),
    ),
)
def test_download_from_github(tmp_path, ds_name, grid_name):
    cache_dir = tmp_path / tutorial._default_cache_dir_name
    ds = tutorial.open_dataset(ds_name, grid_name, cache_dir=cache_dir).load()

    assert cache_dir.is_dir() and len(list(cache_dir.iterdir())) == 1
    assert ds["air"].count() > 0



import operator
from dataclasses import dataclass
from typing import Any, Self, TypeVar

from xdggs.itertools import groupby, identity

T = TypeVar("T")


@dataclass(frozen=True)
class DGGSInfo:
    """Base class for DGGS grid information objects

    Parameters
    ----------
    level : int
        Grid hierarchical level. A higher value corresponds to a finer grid resolution
        with smaller cell areas. The number of cells covering the whole sphere usually
        grows exponentially with increasing level values, ranging from 5-100 cells at
        level 0 to millions or billions of cells at level 10+ (the exact numbers depends
        on the specific grid).
    """

    level: int

    @classmethod
    def from_dict(cls: type[T], mapping: dict[str, Any]) -> T:
        return cls(**mapping)

    def to_dict(self: Self) -> dict[str, Any]:
        return {"level": self.level}

    def cell_ids2geographic(self, cell_ids):
        raise NotImplementedError()

    def geographic2cell_ids(self, lon, lat):
        raise NotImplementedError()

    def cell_boundaries(self, cell_ids, backend="shapely"):
        raise NotImplementedError()

    def zoom_to(self, cell_ids, level: int):
        raise NotImplementedError()


def translate_parameters(mapping, translations):
    def translate(name, value):
        new_name, translator = translations.get(name, (name, identity))

        return new_name, name, translator(value)

    translated = (translate(name, value) for name, value in mapping.items())
    grouped = {
        name: [(old_name, value) for _, old_name, value in group]
        for name, group in groupby(translated, key=operator.itemgetter(0))
    }
    duplicated_parameters = {
        name: group for name, group in grouped.items() if len(group) != 1
    }
    if duplicated_parameters:
        raise ExceptionGroup(
            "received multiple values for parameters",
            [
                ValueError(
                    f"Parameter {name} received multiple values: {sorted(n for n, _ in group)}"
                )
                for name, group in duplicated_parameters.items()
            ],
        )

    params = {
        name: group[0][1] for name, group in grouped.items() if name != "grid_name"
    }
    return params



import xdggs.tutorial  # noqa: F401
from xdggs._version import __version__
from xdggs.accessor import DGGSAccessor  # noqa: F401
from xdggs.grid import DGGSInfo
from xdggs.h3 import H3Index, H3Info
from xdggs.healpix import HealpixIndex, HealpixInfo
from xdggs.index import DGGSIndex, decode

__all__ = [
    "__version__",
    "DGGSInfo",
    "H3Info",
    "HealpixInfo",
    "DGGSIndex",
    "H3Index",
    "HealpixIndex",
    "decode",
]



import ipywidgets
import lonboard
import numpy as np
import pytest
import xarray as xr
from arro3.core import Array, Table
from matplotlib import colormaps

from xdggs import plotting


@pytest.mark.parametrize(
    ["polygons", "arr", "coords", "expected"],
    (
        pytest.param(
            Array.from_numpy(np.array([1, 2])),
            xr.DataArray(
                [-1, 1],
                coords={
                    "cell_ids": ("cells", [0, 1]),
                    "latitude": ("cells", [-5, 10]),
                    "longitude": ("cells", [-60, -50]),
                },
                dims="cells",
            ),
            None,
            Table.from_pydict(
                {
                    "geometry": Array.from_numpy(np.array([1, 2])),
                    "cell_ids": Array.from_numpy(np.array([0, 1])),
                    "data": Array.from_numpy(np.array([-1, 1])),
                    "latitude": Array.from_numpy(np.array([-5, 10])),
                    "longitude": Array.from_numpy(np.array([-60, -50])),
                }
            ),
        ),
        pytest.param(
            Array.from_numpy(np.array([1, 2])),
            xr.DataArray(
                [-1, 1],
                coords={
                    "cell_ids": ("cells", [1, 2]),
                    "latitude": ("cells", [-5, 10]),
                    "longitude": ("cells", [-60, -50]),
                },
                dims="cells",
            ),
            ["latitude"],
            Table.from_pydict(
                {
                    "geometry": Array.from_numpy(np.array([1, 2])),
                    "cell_ids": Array.from_numpy(np.array([1, 2])),
                    "data": Array.from_numpy(np.array([-1, 1])),
                    "latitude": Array.from_numpy(np.array([-5, 10])),
                }
            ),
        ),
        pytest.param(
            Array.from_numpy(np.array([1, 3])),
            xr.DataArray(
                [-1, 1],
                coords={
                    "cell_ids": ("cells", [0, 1]),
                    "latitude": ("cells", [-5, 10]),
                    "longitude": ("cells", [-60, -50]),
                },
                dims="cells",
                name="new_data",
            ),
            ["longitude"],
            Table.from_pydict(
                {
                    "geometry": Array.from_numpy(np.array([1, 3])),
                    "cell_ids": Array.from_numpy(np.array([0, 1])),
                    "new_data": Array.from_numpy(np.array([-1, 1])),
                    "longitude": Array.from_numpy(np.array([-60, -50])),
                }
            ),
        ),
        pytest.param(
            Array.from_numpy(np.array([1, 3])),
            xr.DataArray(
                np.arange(4).reshape((2, 2))[:, 0],
                coords={
                    "cell_ids": ("cells", [0, 1]),
                    "latitude": ("cells", [-5, 10]),
                    "longitude": ("cells", [-60, -50]),
                },
                dims="cells",
                name="new_data",
            ),
            ["longitude"],
            Table.from_pydict(
                {
                    "geometry": Array.from_numpy(np.array([1, 3])),
                    "cell_ids": Array.from_numpy(np.array([0, 1])),
                    "new_data": Array.from_numpy(np.array([0, 2])),
                    "longitude": Array.from_numpy(np.array([-60, -50])),
                }
            ),
            id="non-contiguous",
        ),
    ),
)
def test_create_arrow_table(polygons, arr, coords, expected):
    actual = plotting.create_arrow_table(polygons, arr, coords=coords)

    assert actual == expected


@pytest.mark.parametrize(
    ["var", "center", "expected"],
    (
        pytest.param(
            xr.Variable("cells", np.array([-5, np.nan, -2, 1])),
            None,
            np.array([0, np.nan, 0.5, 1]),
            id="linear-missing_values",
        ),
        pytest.param(
            xr.Variable("cells", np.arange(-5, 2, dtype="float")),
            None,
            np.linspace(0, 1, 7),
            id="linear-manual",
        ),
        pytest.param(
            xr.Variable("cells", np.linspace(0, 10, 5)),
            None,
            np.linspace(0, 1, 5),
            id="linear-linspace",
        ),
        pytest.param(
            xr.Variable("cells", np.linspace(-5, 5, 10)),
            0,
            np.linspace(0, 1, 10),
            id="centered-0",
        ),
        pytest.param(
            xr.Variable("cells", np.linspace(0, 10, 10)),
            5,
            np.linspace(0, 1, 10),
            id="centered-2",
        ),
    ),
)
def test_normalize(var, center, expected):
    actual = plotting.normalize(var, center=center)

    np.testing.assert_allclose(actual, expected)


@pytest.mark.parametrize(
    ["var", "kwargs", "expected"],
    (
        pytest.param(
            xr.Variable("cells", [0, 3]),
            {"center": 2, "colormap": colormaps["viridis"], "alpha": 1},
            np.array([[68, 1, 84], [94, 201, 97]], dtype="uint8"),
        ),
        pytest.param(
            xr.Variable("cells", [-1, 1]),
            {"center": None, "colormap": colormaps["viridis"], "alpha": 0.8},
            np.array([[68, 1, 84, 204], [253, 231, 36, 204]], dtype="uint8"),
        ),
    ),
)
def test_colorize(var, kwargs, expected):
    actual = plotting.colorize(var, **kwargs)

    np.testing.assert_equal(actual, expected)


class TestMapContainer:
    def test_init(self):
        map_ = lonboard.Map(layers=[])
        sliders = ipywidgets.VBox(
            [ipywidgets.IntSlider(min=0, max=10, description="time")]
        )
        obj = xr.DataArray([[0, 1], [2, 3]], dims=["time", "cells"])
        colorize_kwargs = {"a": 1, "b": 2}

        container = plotting.MapContainer(
            dimension_sliders=sliders,
            map=map_,
            obj=obj,
            colorize_kwargs=colorize_kwargs,
        )

        assert container.map == map_
        xr.testing.assert_equal(container.obj, obj)
        assert container.dimension_sliders == sliders
        assert container.colorize_kwargs == colorize_kwargs

    def test_render(self):
        map_ = lonboard.Map(layers=[])
        sliders = ipywidgets.VBox(
            [ipywidgets.IntSlider(min=0, max=10, description="time")]
        )
        obj = xr.DataArray([[0, 1], [2, 3]], dims=["time", "cells"])
        colorize_kwargs = {"a": 1, "b": 2}

        container = plotting.MapContainer(
            dimension_sliders=sliders,
            map=map_,
            obj=obj,
            colorize_kwargs=colorize_kwargs,
        )
        rendered = container.render()

        assert isinstance(rendered, ipywidgets.VBox)


@pytest.mark.parametrize(
    ["arr", "expected_type"],
    (
        pytest.param(
            xr.DataArray(
                [0, 1], coords={"cell_ids": ("cells", [10, 26])}, dims="cells"
            ).dggs.decode(
                {"grid_name": "healpix", "level": 1, "indexing_scheme": "nested"}
            ),
            lonboard.Map,
            id="1d",
        ),
        pytest.param(
            xr.DataArray(
                [[0, 1], [2, 3]],
                coords={"cell_ids": ("cells", [10, 26])},
                dims=["time", "cells"],
            ).dggs.decode(
                {"grid_name": "healpix", "level": 1, "indexing_scheme": "nested"}
            ),
            ipywidgets.VBox,
            id="2d",
        ),
    ),
)
def test_explore(arr, expected_type):
    actual = arr.dggs.explore()

    assert isinstance(actual, expected_type)



from dataclasses import dataclass
from functools import partial
from typing import Any

import ipywidgets
import numpy as np
import xarray as xr
from lonboard import Map


def on_slider_change(change, container):
    owner = change["owner"]
    dim = owner.description

    indexers = {
        slider.description: slider.value
        for slider in container.dimension_sliders.children
        if slider.description != dim
    } | {dim: change["new"]}
    new_slice = container.obj.isel(indexers)

    colors = colorize(new_slice.variable, **container.colorize_kwargs)

    layer = container.map.layers[0]
    layer.get_fill_color = colors


@dataclass
class MapContainer:
    """container for the map, any control widgets and the data object"""

    dimension_sliders: ipywidgets.VBox
    map: Map
    obj: xr.DataArray

    colorize_kwargs: dict[str, Any]

    def render(self):
        # add any additional control widgets here
        control_box = ipywidgets.HBox([self.dimension_sliders])

        return ipywidgets.VBox([self.map, control_box])


def create_arrow_table(polygons, arr, coords=None):
    from arro3.core import Array, ChunkedArray, Schema, Table

    if coords is None:
        coords = ["latitude", "longitude"]

    array = Array.from_arrow(polygons)
    name = arr.name or "data"
    arrow_arrays = {
        "geometry": array,
        "cell_ids": ChunkedArray([Array.from_numpy(arr.coords["cell_ids"])]),
        name: ChunkedArray([Array.from_numpy(np.ascontiguousarray(arr.data))]),
    } | {
        coord: ChunkedArray([Array.from_numpy(arr.coords[coord].data)])
        for coord in coords
        if coord in arr.coords
    }

    fields = [array.field.with_name(name) for name, array in arrow_arrays.items()]
    schema = Schema(fields)

    return Table.from_arrays(list(arrow_arrays.values()), schema=schema)


def normalize(var, center=None):
    from matplotlib.colors import CenteredNorm, Normalize

    if center is None:
        vmin = var.min(skipna=True)
        vmax = var.max(skipna=True)
        normalizer = Normalize(vmin=vmin, vmax=vmax)
    else:
        halfrange = np.abs(var - center).max(skipna=True)
        normalizer = CenteredNorm(vcenter=center, halfrange=halfrange)

    return normalizer(var.data)


def colorize(var, *, center, colormap, alpha):
    from lonboard.colormap import apply_continuous_cmap

    normalized_data = normalize(var, center=center)

    return apply_continuous_cmap(normalized_data, colormap, alpha=alpha)


def explore(
    arr,
    cmap="viridis",
    center=None,
    alpha=None,
    coords=None,
):
    import lonboard
    from lonboard import SolidPolygonLayer
    from matplotlib import colormaps

    # guaranteed to be 1D
    cell_id_coord = arr.dggs.coord
    [cell_dim] = cell_id_coord.dims

    cell_ids = cell_id_coord.data
    grid_info = arr.dggs.grid_info

    polygons = grid_info.cell_boundaries(cell_ids, backend="geoarrow")

    initial_indexers = {dim: 0 for dim in arr.dims if dim != cell_dim}
    initial_arr = arr.isel(initial_indexers)

    colormap = colormaps[cmap] if isinstance(cmap, str) else cmap
    colors = colorize(initial_arr, center=center, alpha=alpha, colormap=colormap)

    table = create_arrow_table(polygons, initial_arr, coords=coords)
    layer = SolidPolygonLayer(table=table, filled=True, get_fill_color=colors)

    map_ = lonboard.Map(layer)

    if not initial_indexers:
        # 1D data
        return map_

    sliders = ipywidgets.VBox(
        [
            ipywidgets.IntSlider(min=0, max=arr.sizes[dim] - 1, description=dim)
            for dim in arr.dims
            if dim != cell_dim
        ]
    )

    container = MapContainer(
        sliders,
        map_,
        arr,
        colorize_kwargs={"alpha": alpha, "center": center, "colormap": colormap},
    )

    # connect slider with map
    for slider in sliders.children:
        slider.observe(partial(on_slider_change, container=container), names="value")

    return container.render()



GRID_REGISTRY = {}


def register_dggs(name):
    def inner(cls):
        GRID_REGISTRY[name] = cls
        return cls

    return inner


def _extract_cell_id_variable(variables):
    # TODO: only one variable supported (raise otherwise)
    name, var = next(iter(variables.items()))

    # TODO: only 1-d variable supported (raise otherwise)
    dim = next(iter(var.dims))

    return name, var, dim



import pytest

from xdggs.tests import matchers


class TestMatch:
    def test_construct_simple(self):
        # simple
        matchers.Match(ValueError)
        matchers.Match((ValueError, NameError))
        matchers.Match(ExceptionGroup)
        with pytest.raises(
            TypeError, match="exception type must be one or more exceptions"
        ):
            matchers.Match(int)

        # with match string
        matchers.Match(TypeError, match="pattern")
        matchers.Match(ExceptionGroup, match="pattern")
        with pytest.raises(TypeError):
            matchers.Match(ValueError, match=int)

        # with submatchers
        with pytest.raises(TypeError):
            matchers.Match(ValueError, submatchers=[matchers.Match(ValueError)])

    @pytest.mark.parametrize(
        ["mapping", "args", "kwargs"],
        (
            ({"exceptions": ValueError}, (ValueError,), {}),
            (
                {"exceptions": ValueError, "match": "abc"},
                (ValueError,),
                {"match": "abc"},
            ),
        ),
    )
    def test_from_dict(self, mapping, args, kwargs):
        actual = matchers.Match.from_dict(mapping)
        expected = matchers.Match(*args, **kwargs)
        assert actual == expected

    @pytest.mark.parametrize(
        ["exc", "match", "expected"],
        (
            pytest.param(
                ValueError("e"), matchers.Match(ValueError), True, id="exc-match-wo pat"
            ),
            pytest.param(
                ValueError("error message"),
                matchers.Match(ValueError, match="message"),
                True,
                id="exc-match-w pat",
            ),
            pytest.param(
                ValueError("error message"),
                matchers.Match(ValueError, match="abc"),
                False,
                id="exc-no match-w pat",
            ),
            pytest.param(
                ExceptionGroup("eg", [ValueError("error")]),
                matchers.Match(ExceptionGroup),
                True,
                id="eg-match-wo pat-without submatchers",
            ),
            pytest.param(
                ExceptionGroup("error group", [ValueError("error")]),
                matchers.Match(ExceptionGroup, match="err"),
                True,
                id="eg-match-w pat-without submatchers",
            ),
            pytest.param(
                ExceptionGroup("eg", [ValueError("error")]),
                matchers.Match(ExceptionGroup, match="abc"),
                False,
                id="eg-no match-w pat-without submatchers",
            ),
            pytest.param(
                ExceptionGroup("eg", [ValueError("error")]),
                matchers.Match(
                    ExceptionGroup, submatchers=[matchers.Match(ValueError)]
                ),
                True,
                id="eg-match-wo pat-with submatchers-wo subpat",
            ),
            pytest.param(
                ExceptionGroup("eg", [ValueError("error")]),
                matchers.Match(ExceptionGroup, submatchers=[matchers.Match(TypeError)]),
                False,
                id="eg-no match-wo pat-with submatchers-wo subpat",
            ),
            pytest.param(
                ExceptionGroup("eg", [ValueError("error")]),
                matchers.Match(
                    ExceptionGroup,
                    submatchers=[matchers.Match(ValueError, match="err")],
                ),
                True,
                id="eg-match-w pat-with submatchers-wo subpat",
            ),
            pytest.param(
                ExceptionGroup("eg", [ValueError("error")]),
                matchers.Match(
                    ExceptionGroup,
                    submatchers=[matchers.Match(ValueError, match="abc")],
                ),
                False,
                id="eg-no match-w pat-with submatchers-wo subpat",
            ),
        ),
    )
    def test_matches(self, exc, match, expected):
        assert match.matches(exc) == expected


def test_assert_exceptions_equal():
    actual = ValueError("error message")
    expected = ValueError("error message")

    matchers.assert_exceptions_equal(actual, expected)

    try:
        raise ValueError("error message")
    except ValueError as e:
        actual = e
    expected = ValueError("error message")

    matchers.assert_exceptions_equal(actual, expected)

    actual = ValueError("error message")
    expected = TypeError("error message")
    with pytest.raises(AssertionError):
        matchers.assert_exceptions_equal(actual, expected)

    actual = ValueError("error message1")
    expected = ValueError("error message2")
    with pytest.raises(AssertionError):
        matchers.assert_exceptions_equal(actual, expected)

    actual = ExceptionGroup("group message1", [ValueError("error message")])
    expected = ExceptionGroup("group message2", [ValueError("error message")])
    with pytest.raises(AssertionError):
        matchers.assert_exceptions_equal(actual, expected)

    actual = ExceptionGroup("group message", [ValueError("error message1")])
    expected = ExceptionGroup("group message", [ValueError("error message2")])
    with pytest.raises(AssertionError):
        matchers.assert_exceptions_equal(actual, expected)

    actual = ExceptionGroup("group message", [ValueError("error message")])
    expected = ExceptionGroup("group message", [TypeError("error message")])
    with pytest.raises(AssertionError):
        matchers.assert_exceptions_equal(actual, expected)



import itertools

import hypothesis.extra.numpy as npst
import hypothesis.strategies as st
import numpy as np
import pytest
import shapely
import shapely.testing
import xarray as xr
import xarray.testing.strategies as xrst
from hypothesis import given
from xarray.core.indexes import PandasIndex

from xdggs import healpix
from xdggs.tests import assert_exceptions_equal, geoarrow_to_shapely


# namespace class
class strategies:
    invalid_levels = st.integers(max_value=-1) | st.integers(min_value=30)
    levels = st.integers(min_value=0, max_value=29)
    # TODO: add back `"unique"` once that is supported
    indexing_schemes = st.sampled_from(["nested", "ring"])
    invalid_indexing_schemes = st.text().filter(lambda x: x not in ["nested", "ring"])

    dims = xrst.names()

    @classmethod
    def grid_mappings(cls):
        strategies = {
            "resolution": cls.levels,
            "nside": cls.levels.map(lambda n: 2**n),
            "depth": cls.levels,
            "level": cls.levels,
            "order": cls.levels,
            "indexing_scheme": cls.indexing_schemes,
            "nest": st.booleans(),
        }

        names = {
            "level": st.sampled_from(
                ["resolution", "nside", "depth", "level", "order"]
            ),
            "indexing_scheme": st.sampled_from(["indexing_scheme", "nest"]),
        }

        return st.builds(lambda **x: list(x.values()), **names).flatmap(
            lambda params: st.builds(dict, **{p: strategies[p] for p in params})
        )

    def cell_ids(max_value=None, dtypes=None):
        if dtypes is None:
            dtypes = st.sampled_from(["int32", "int64", "uint32", "uint64"])
        shapes = npst.array_shapes(min_dims=1, max_dims=1)

        return npst.arrays(
            dtypes,
            shapes,
            elements={"min_value": 0, "max_value": max_value},
            unique=True,
            fill=st.nothing(),
        )

    options = st.just({})

    def grids(
        levels=levels,
        indexing_schemes=indexing_schemes,
    ):
        return st.builds(
            healpix.HealpixInfo,
            level=levels,
            indexing_scheme=indexing_schemes,
        )

    @classmethod
    def grid_and_cell_ids(
        cls,
        levels=levels,
        indexing_schemes=indexing_schemes,
        dtypes=None,
    ):
        cell_levels = st.shared(levels, key="common-levels")
        grid_levels = st.shared(levels, key="common-levels")
        cell_ids_ = cell_levels.flatmap(
            lambda level: cls.cell_ids(max_value=12 * 4**level - 1, dtypes=dtypes)
        )
        grids_ = cls.grids(
            levels=grid_levels,
            indexing_schemes=indexing_schemes,
        )

        return cell_ids_, grids_


options = [{}]
variable_names = ["cell_ids", "zonal_ids", "zone_ids"]
variables = [
    xr.Variable(
        "cells",
        np.array([3]),
        {
            "grid_name": "healpix",
            "level": 0,
            "indexing_scheme": "nested",
        },
    ),
    xr.Variable(
        "zones",
        np.array([3]),
        {
            "grid_name": "healpix",
            "level": 0,
            "indexing_scheme": "ring",
        },
    ),
    xr.Variable(
        "cells",
        np.array([5, 11, 21]),
        {
            "grid_name": "healpix",
            "level": 1,
            "indexing_scheme": "nested",
        },
    ),
    xr.Variable(
        "zones",
        np.array([54, 70, 82, 91]),
        {
            "grid_name": "healpix",
            "level": 3,
            "indexing_scheme": "nested",
        },
    ),
]
variable_combinations = list(itertools.product(variables, repeat=2))


class TestHealpixInfo:
    @given(strategies.invalid_levels)
    def test_init_invalid_levels(self, level):
        with pytest.raises(
            ValueError, match="level must be an integer in the range of"
        ):
            healpix.HealpixInfo(level=level)

    @given(strategies.invalid_indexing_schemes)
    def test_init_invalid_indexing_scheme(self, indexing_scheme):
        with pytest.raises(ValueError, match="indexing scheme must be one of"):
            healpix.HealpixInfo(
                level=0,
                indexing_scheme=indexing_scheme,
            )

    @given(strategies.levels, strategies.indexing_schemes)
    def test_init(self, level, indexing_scheme):
        grid = healpix.HealpixInfo(level=level, indexing_scheme=indexing_scheme)

        assert grid.level == level
        assert grid.indexing_scheme == indexing_scheme

    @given(strategies.levels)
    def test_nside(self, level):
        grid = healpix.HealpixInfo(level=level)

        assert grid.nside == 2**level

    @given(strategies.indexing_schemes)
    def test_nest(self, indexing_scheme):
        grid = healpix.HealpixInfo(level=1, indexing_scheme=indexing_scheme)
        if indexing_scheme not in {"nested", "ring"}:
            with pytest.raises(
                ValueError, match="cannot convert indexing scheme .* to `nest`"
            ):
                grid.nest
            return

        expected = indexing_scheme == "nested"

        assert grid.nest == expected

    @given(strategies.grid_mappings())
    def test_from_dict(self, mapping) -> None:
        healpix.HealpixInfo.from_dict(mapping)

    @given(strategies.levels, strategies.indexing_schemes)
    def test_to_dict(self, level, indexing_scheme) -> None:
        grid = healpix.HealpixInfo(level=level, indexing_scheme=indexing_scheme)
        actual = grid.to_dict()

        assert set(actual) == {"grid_name", "level", "indexing_scheme"}
        assert actual["grid_name"] == "healpix"
        assert actual["level"] == level
        assert actual["indexing_scheme"] == indexing_scheme

    @given(strategies.levels, strategies.indexing_schemes)
    def test_roundtrip(self, level, indexing_scheme):
        mapping = {
            "grid_name": "healpix",
            "level": level,
            "indexing_scheme": indexing_scheme,
        }

        grid = healpix.HealpixInfo.from_dict(mapping)
        roundtripped = grid.to_dict()

        assert roundtripped == mapping

    @pytest.mark.parametrize(
        ["params", "cell_ids", "expected_coords"],
        (
            (
                {"level": 0, "indexing_scheme": "nested"},
                np.array([2]),
                np.array(
                    [
                        [-135.0, 0.0],
                        [-90.0, 41.8103149],
                        [-135.0, 90.0],
                        [-180.0, 41.8103149],
                    ]
                ),
            ),
            (
                {"level": 2, "indexing_scheme": "ring"},
                np.array([12, 54]),
                np.array(
                    [
                        [
                            [22.5, 41.8103149],
                            [30.0, 54.3409123],
                            [0.0, 66.44353569],
                            [0.0, 54.3409123],
                        ],
                        [
                            [-45.0, 19.47122063],
                            [-33.75, 30.0],
                            [-45.0, 41.8103149],
                            [-56.25, 30.0],
                        ],
                    ]
                ),
            ),
            (
                {"level": 3, "indexing_scheme": "nested"},
                np.array([293, 17]),
                np.array(
                    [
                        [
                            [-5.625, -4.78019185],
                            [0.0, 0.0],
                            [-5.625, 4.78019185],
                            [-11.25, 0.0],
                        ],
                        [
                            [73.125, 24.62431835],
                            [78.75, 30.0],
                            [73.125, 35.68533471],
                            [67.5, 30.0],
                        ],
                    ]
                ),
            ),
            (
                {"level": 2, "indexing_scheme": "nested"},
                np.array([79]),
                np.array(
                    [
                        [0.0, 19.47122063],
                        [11.25, 30],
                        [0.0, 41.8103149],
                        [-11.25, 30],
                    ]
                ),
            ),
        ),
    )
    @pytest.mark.parametrize("backend", ["shapely", "geoarrow"])
    def test_cell_boundaries(self, params, cell_ids, backend, expected_coords):
        grid = healpix.HealpixInfo.from_dict(params)

        actual = grid.cell_boundaries(cell_ids, backend=backend)

        backends = {
            "shapely": lambda arr: arr,
            "geoarrow": geoarrow_to_shapely,
        }
        converter = backends[backend]
        expected = shapely.polygons(expected_coords)

        shapely.testing.assert_geometries_equal(converter(actual), expected)

    @given(
        *strategies.grid_and_cell_ids(
            # a dtype casting bug in the valid range check of `cdshealpix`
            # causes this test to fail for large levels
            levels=st.integers(min_value=0, max_value=10),
            indexing_schemes=st.sampled_from(["nested", "ring"]),
            dtypes=st.sampled_from(["int64"]),
        )
    )
    def test_cell_center_roundtrip(self, cell_ids, grid) -> None:
        centers = grid.cell_ids2geographic(cell_ids)

        roundtripped = grid.geographic2cell_ids(lat=centers[1], lon=centers[0])

        np.testing.assert_equal(roundtripped, cell_ids)

    @pytest.mark.parametrize(
        ["cell_ids", "level", "indexing_scheme", "expected"],
        (
            pytest.param(
                np.array([3]),
                1,
                "ring",
                (np.array([315.0]), np.array([66.44353569089877])),
            ),
            pytest.param(
                np.array([5, 11, 21]),
                3,
                "nested",
                (
                    np.array([61.875, 33.75, 84.375]),
                    np.array([19.47122063, 24.62431835, 41.8103149]),
                ),
            ),
        ),
    )
    def test_cell_ids2geographic(
        self, cell_ids, level, indexing_scheme, expected
    ) -> None:
        grid = healpix.HealpixInfo(level=level, indexing_scheme=indexing_scheme)

        actual_lon, actual_lat = grid.cell_ids2geographic(cell_ids)

        np.testing.assert_allclose(actual_lon, expected[0])
        np.testing.assert_allclose(actual_lat, expected[1])

    @pytest.mark.parametrize(
        ["cell_centers", "level", "indexing_scheme", "expected"],
        (
            pytest.param(
                np.array([[315.0, 66.44353569089877]]),
                1,
                "ring",
                np.array([3]),
            ),
            pytest.param(
                np.array(
                    [[61.875, 19.47122063], [33.75, 24.62431835], [84.375, 41.8103149]]
                ),
                3,
                "nested",
                np.array([5, 11, 21]),
            ),
        ),
    )
    def test_geographic2cell_ids(
        self, cell_centers, level, indexing_scheme, expected
    ) -> None:
        grid = healpix.HealpixInfo(level=level, indexing_scheme=indexing_scheme)

        actual = grid.geographic2cell_ids(
            lon=cell_centers[:, 0], lat=cell_centers[:, 1]
        )

        np.testing.assert_equal(actual, expected)

    @pytest.mark.parametrize(
        ["level", "cell_ids", "new_level", "expected"],
        (
            pytest.param(
                1,
                np.array([0, 4, 8, 12, 16]),
                0,
                np.array([0, 1, 2, 3, 4]),
                id="level1-parents",
            ),
            pytest.param(
                1,
                np.array([0, 1, 2, 3]),
                2,
                np.array(
                    [[0, 1, 2, 3], [4, 5, 6, 7], [8, 9, 10, 11], [12, 13, 14, 15]]
                ),
                id="level1-children",
            ),
            pytest.param(
                1,
                np.array([0, 4]),
                3,
                np.stack([np.arange(16), 4 * 4**2 + np.arange(16)]),
                id="level1-grandchildren",
            ),
        ),
    )
    def test_zoom_to(self, level, cell_ids, new_level, expected):
        grid = healpix.HealpixInfo(level=level, indexing_scheme="nested")

        actual = grid.zoom_to(cell_ids, level=new_level)

        np.testing.assert_equal(actual, expected)

    def test_zoom_to_ring(self):
        cell_ids = np.array([1, 2, 3])
        grid = healpix.HealpixInfo(level=1, indexing_scheme="ring")

        with pytest.raises(ValueError, match="Scaling does not make sense.*'ring'.*"):
            grid.zoom_to(cell_ids, level=0)


@pytest.mark.parametrize(
    ["mapping", "expected"],
    (
        pytest.param(
            {"resolution": 10, "indexing_scheme": "nested"},
            {"level": 10, "indexing_scheme": "nested"},
            id="no_translation",
        ),
        pytest.param(
            {
                "level": 10,
                "indexing_scheme": "nested",
                "grid_name": "healpix",
            },
            {"level": 10, "indexing_scheme": "nested"},
            id="no_translation-grid_name",
        ),
        pytest.param(
            {"nside": 1024, "indexing_scheme": "nested"},
            {"level": 10, "indexing_scheme": "nested"},
            id="nside-alone",
        ),
        pytest.param(
            {
                "nside": 1024,
                "level": 10,
                "indexing_scheme": "nested",
            },
            ExceptionGroup(
                "received multiple values for parameters",
                [
                    ValueError(
                        "Parameter level received multiple values: ['level', 'nside']"
                    )
                ],
            ),
            id="nside-duplicated",
        ),
        pytest.param(
            {
                "level": 10,
                "indexing_scheme": "nested",
                "nest": True,
            },
            ExceptionGroup(
                "received multiple values for parameters",
                [
                    ValueError(
                        "Parameter indexing_scheme received multiple values: ['indexing_scheme', 'nest']"
                    ),
                ],
            ),
            id="indexing_scheme-duplicated",
        ),
        pytest.param(
            {
                "nside": 1024,
                "level": 10,
                "indexing_scheme": "nested",
                "nest": True,
            },
            ExceptionGroup(
                "received multiple values for parameters",
                [
                    ValueError(
                        "Parameter indexing_scheme received multiple values: ['indexing_scheme', 'nest']"
                    ),
                    ValueError(
                        "Parameter level received multiple values: ['level', 'nside']"
                    ),
                ],
            ),
            id="multiple_params-duplicated",
        ),
    ),
)
def test_healpix_info_from_dict(mapping, expected) -> None:
    if isinstance(expected, ExceptionGroup):
        with pytest.raises(type(expected), match=expected.args[0]) as actual:
            healpix.HealpixInfo.from_dict(mapping)
        assert_exceptions_equal(actual.value, expected)
        return
    actual = healpix.HealpixInfo.from_dict(mapping)
    assert actual == healpix.HealpixInfo(**expected)


class TestHealpixIndex:
    @given(strategies.cell_ids(), strategies.dims, strategies.grids())
    def test_init(self, cell_ids, dim, grid) -> None:
        index = healpix.HealpixIndex(cell_ids, dim, grid)

        assert index._grid == grid
        assert index._dim == dim
        assert index._pd_index.dim == dim

        np.testing.assert_equal(index._pd_index.index.values, cell_ids)

    @given(strategies.grids())
    def test_grid(self, grid):
        index = healpix.HealpixIndex([0], dim="cells", grid_info=grid)

        assert index.grid_info is grid


@pytest.mark.parametrize("options", options)
@pytest.mark.parametrize("variable", variables)
@pytest.mark.parametrize("variable_name", variable_names)
def test_from_variables(variable_name, variable, options) -> None:
    expected_level = variable.attrs["level"]
    expected_scheme = variable.attrs["indexing_scheme"]

    variables = {variable_name: variable}

    index = healpix.HealpixIndex.from_variables(variables, options=options)

    assert index._grid.level == expected_level
    assert index._grid.indexing_scheme == expected_scheme

    assert (index._dim,) == variable.dims
    np.testing.assert_equal(index._pd_index.index.values, variable.data)


@pytest.mark.parametrize(["old_variable", "new_variable"], variable_combinations)
def test_replace(old_variable, new_variable) -> None:
    grid = healpix.HealpixInfo.from_dict(old_variable.attrs)

    index = healpix.HealpixIndex(
        cell_ids=old_variable.data,
        dim=old_variable.dims[0],
        grid_info=grid,
    )

    new_pandas_index = PandasIndex.from_variables(
        {"cell_ids": new_variable}, options={}
    )

    new_index = index._replace(new_pandas_index)

    assert new_index._dim == index._dim
    assert new_index._pd_index == new_pandas_index
    assert index._grid == grid


@pytest.mark.parametrize("max_width", [20, 50, 80, 120])
@pytest.mark.parametrize("level", [0, 1, 3])
def test_repr_inline(level, max_width) -> None:
    grid_info = healpix.HealpixInfo(level=level, indexing_scheme="nested")
    index = healpix.HealpixIndex(cell_ids=[0], dim="cells", grid_info=grid_info)

    actual = index._repr_inline_(max_width)

    assert f"level={level}" in actual
    # ignore max_width for now
    # assert len(actual) <= max_width



from __future__ import annotations

import os
import pathlib
from typing import TYPE_CHECKING

import pooch
from xarray import open_dataset as _open_dataset

if TYPE_CHECKING:
    from xarray.backends.api import T_Engine

_default_cache_dir_name = "xdggs_tutorial_data"
base_url = "https://github.com/xdggs/xdggs-data"
version = "main"

external_urls = {}  # type: dict
file_formats = {
    "air_temperature": 4,
}


def _construct_cache_dir(path):
    if isinstance(path, os.PathLike):
        path = os.fspath(path)
    elif path is None:
        path = pooch.os_cache(_default_cache_dir_name)

    return path


def _check_netcdf_engine_installed(name):
    version = file_formats.get(name)
    if version == 3:
        try:
            import scipy  # noqa
        except ImportError:
            try:
                import netCDF4  # noqa
            except ImportError as err:
                raise ImportError(
                    f"opening tutorial dataset {name} requires either scipy or "
                    "netCDF4 to be installed."
                ) from err
    if version == 4:
        try:
            import h5netcdf  # noqa
        except ImportError:
            try:
                import netCDF4  # noqa
            except ImportError as err:
                raise ImportError(
                    f"opening tutorial dataset {name} requires either h5netcdf "
                    "or netCDF4 to be installed."
                ) from err


def open_dataset(
    name: str,
    grid_name: str,
    *,
    cache: bool = True,
    cache_dir: None | str | os.PathLike = None,
    engine: T_Engine = None,
    **kws,
):
    """
    Open a dataset from the online repository (requires internet).

    If a local copy is found then always use that to avoid network traffic.

    Available datasets (available grid names in parentheses):

    * ``"air_temperature"`` (``h3``, ``healpix``): NCEP reanalysis subset.

    Parameters
    ----------
    name : str
        Name of the file containing the dataset.
        e.g. 'air_temperature'
    grid_name : str
        Name of the grid file.
    cache_dir : path-like, optional
        The directory in which to search for and write cached data.
    cache : bool, optional
        If True, then cache data locally for use on subsequent calls
    **kws : dict, optional
        Passed to xarray.open_dataset

    See Also
    --------
    xarray.tutorial.open_dataset
    """
    import xdggs

    logger = pooch.get_logger()
    logger.setLevel("WARNING")

    cache_dir = _construct_cache_dir(cache_dir)
    if name in external_urls:
        url = external_urls[name]
    else:
        path = pathlib.Path(grid_name)
        if not path.suffix:
            # process the name
            default_extension = ".nc"
            if engine is None:
                _check_netcdf_engine_installed(grid_name)
            path = path.with_suffix(default_extension)

        url = f"{base_url}/raw/{version}/{name}/{path.name}"

    headers = {"User-Agent": f"xdggs/{xdggs.__version__}"}

    # retrieve the file
    downloader = pooch.HTTPDownloader(headers=headers)
    filepath = pooch.retrieve(
        url=url, known_hash=None, path=cache_dir, downloader=downloader
    )
    ds = _open_dataset(filepath, engine=engine, **kws)
    if not cache:
        ds = ds.load()
        pathlib.Path(filepath).unlink()

    return ds



from collections.abc import Hashable, Mapping
from typing import Any, Union

import numpy as np
import xarray as xr
from xarray.indexes import Index, PandasIndex

from xdggs.grid import DGGSInfo
from xdggs.utils import GRID_REGISTRY, _extract_cell_id_variable


def decode(ds, grid_info=None, *, name="cell_ids"):
    """
    decode grid parameters and create a DGGS index

    Parameters
    ----------
    ds : xarray.Dataset
        The input dataset. Must contain a coordinate for the cell ids with at
        least the attributes `grid_name` and `level`.
    grid_info : dict or DGGSInfo, optional
        Override the grid parameters on the dataset. Useful to set attributes on
        the dataset.
    name : str, default: "cell_ids"
        The name of the coordinate containing the cell ids.

    Returns
    -------
    decoded : xarray.DataArray or xarray.Dataset
        The input dataset with a DGGS index on the cell id coordinate.

    See Also
    --------
    xarray.Dataset.dggs.decode
    xarray.DataArray.dggs.decode
    """
    return ds.dggs.decode(name=name, grid_info=grid_info)


class DGGSIndex(Index):
    _dim: str
    _pd_index: PandasIndex

    def __init__(self, cell_ids: Any | PandasIndex, dim: str, grid_info: DGGSInfo):
        self._dim = dim

        if isinstance(cell_ids, PandasIndex):
            self._pd_index = cell_ids
        else:
            self._pd_index = PandasIndex(cell_ids, dim)

        self._grid = grid_info

    @classmethod
    def from_variables(
        cls: type["DGGSIndex"],
        variables: Mapping[Any, xr.Variable],
        *,
        options: Mapping[str, Any],
    ) -> "DGGSIndex":
        name, var, _ = _extract_cell_id_variable(variables)

        grid_name = var.attrs["grid_name"]
        cls = GRID_REGISTRY.get(grid_name)
        if cls is None:
            raise ValueError(f"unknown DGGS grid name: {grid_name}")

        index = cls.from_variables(variables, options=options)
        index._pd_index.index.name = name

        return index

    def create_variables(
        self, variables: Mapping[Any, xr.Variable] | None = None
    ) -> dict[Hashable, xr.Variable]:
        return self._pd_index.create_variables(variables)

    def isel(
        self: "DGGSIndex", indexers: Mapping[Any, int | np.ndarray | xr.Variable]
    ) -> Union["DGGSIndex", None]:
        new_pd_index = self._pd_index.isel(indexers)
        if new_pd_index is not None:
            return self._replace(new_pd_index)
        else:
            return None

    def sel(self, labels, method=None, tolerance=None):
        if method == "nearest":
            raise ValueError("finding nearest grid cell has no meaning")
        return self._pd_index.sel(labels, method=method, tolerance=tolerance)

    def _replace(self, new_pd_index: PandasIndex):
        raise NotImplementedError()

    def cell_centers(self) -> tuple[np.ndarray, np.ndarray]:
        return self._grid.cell_ids2geographic(self._pd_index.index.values)

    def cell_boundaries(self) -> np.ndarray:
        return self.grid_info.cell_boundaries(self._pd_index.index.values)

    def zoom_to(self, level: int) -> np.ndarray:
        return self._grid.zoom_to(self._pd_index.index.values, level=level)

    @property
    def grid_info(self) -> DGGSInfo:
        return self._grid



import pytest
import xarray as xr

from xdggs import index


@pytest.fixture
def dggs_example():
    return xr.Dataset(
        coords={"cell_ids": ("cells", [0, 1], {"grid_name": "test", "level": 2})}
    )


def test_create_index(dggs_example):
    # TODO: improve unknown index message
    with pytest.raises(ValueError, match="test"):
        dggs_example.set_xindex("cell_ids", index.DGGSIndex)


def test_decode(dggs_example):
    # TODO: improve unknown index message
    with pytest.raises(ValueError, match="test"):
        dggs_example.pipe(index.decode)


def test_decode_indexed(dggs_example):
    with pytest.raises(ValueError, match="test"):
        dggs_example.set_xindex("cell_ids").pipe(index.decode)



import json
from collections.abc import Mapping
from dataclasses import dataclass
from typing import Any, ClassVar, Literal, Self, TypeVar

import cdshealpix.nested
import cdshealpix.ring
import numpy as np
import xarray as xr
from xarray.indexes import PandasIndex

from xdggs.grid import DGGSInfo, translate_parameters
from xdggs.index import DGGSIndex
from xdggs.itertools import identity
from xdggs.utils import _extract_cell_id_variable, register_dggs

T = TypeVar("T")


def polygons_shapely(vertices):
    import shapely

    return shapely.polygons(vertices)


def polygons_geoarrow(vertices):
    import pyproj
    from arro3.core import list_array

    polygon_vertices = np.concatenate([vertices, vertices[:, :1, :]], axis=1)
    crs = pyproj.CRS.from_epsg(4326)

    # construct geoarrow arrays
    coords = np.reshape(polygon_vertices, (-1, 2))
    coords_per_pixel = polygon_vertices.shape[1]
    geom_offsets = np.arange(vertices.shape[0] + 1, dtype="int32")
    ring_offsets = geom_offsets * coords_per_pixel

    polygon_array = list_array(geom_offsets, list_array(ring_offsets, coords))

    # We need to tag the array with extension metadata (`geoarrow.polygon`) so that Lonboard knows that this is a geospatial column.
    polygon_array_with_geo_meta = polygon_array.cast(
        polygon_array.field.with_metadata(
            {
                "ARROW:extension:name": "geoarrow.polygon",
                "ARROW:extension:metadata": json.dumps(
                    {"crs": crs.to_json_dict(), "edges": "spherical"}
                ),
            }
        )
    )
    return polygon_array_with_geo_meta


def center_around_prime_meridian(lon, lat):
    # three tasks:
    # - center around the prime meridian (map to a range of [-180, 180])
    # - replace the longitude of points at the poles with the median
    #   of longitude of the other vertices
    # - cells that cross the dateline should have longitudes around 180

    # center around prime meridian
    recentered = (lon + 180) % 360 - 180

    # replace lon of pole with the median of the remaining vertices
    contains_poles = np.isin(lat, np.array([-90, 90]))
    pole_cells = np.any(contains_poles, axis=-1)
    recentered[contains_poles] = np.median(
        np.reshape(
            recentered[pole_cells[:, None] & np.logical_not(contains_poles)], (-1, 3)
        ),
        axis=-1,
    )

    # keep cells that cross the dateline centered around 180
    polygons_to_fix = np.any(recentered < -100, axis=-1) & np.any(
        recentered > 100, axis=-1
    )
    result = np.where(
        polygons_to_fix[:, None] & (recentered < 0), recentered + 360, recentered
    )

    return result


@dataclass(frozen=True)
class HealpixInfo(DGGSInfo):
    """
    Grid information container for healpix grids.

    Parameters
    ----------
    level : int
        Grid hierarchical level. A higher value corresponds to a finer grid resolution
        with smaller cell areas. The number of cells covering the whole sphere usually
        grows exponentially with increasing level values, ranging from 5-100 cells at
        level 0 to millions or billions of cells at level 10+ (the exact numbers depends
        on the specific grid).
    indexing_scheme : {"nested", "ring", "unique"}, default: "nested"
        The indexing scheme of the healpix grid.

        .. warning::
            Note that ``"unique"`` is currently not supported as the underlying library
            (:doc:`cdshealpix <cdshealpix-python:index>`) does not support it.
    """

    level: int
    """int : The hierarchical level of the grid"""

    indexing_scheme: Literal["nested", "ring", "unique"] = "nested"
    """int : The indexing scheme of the grid"""

    valid_parameters: ClassVar[dict[str, Any]] = {
        "level": range(0, 29 + 1),
        "indexing_scheme": ["nested", "ring", "unique"],
    }

    def __post_init__(self):
        if self.level not in self.valid_parameters["level"]:
            raise ValueError("level must be an integer in the range of [0, 29]")

        if self.indexing_scheme not in self.valid_parameters["indexing_scheme"]:
            raise ValueError(
                f"indexing scheme must be one of {self.valid_parameters['indexing_scheme']}"
            )
        elif self.indexing_scheme == "unique":
            raise ValueError("the indexing scheme `unique` is currently not supported")

    @property
    def nside(self: Self) -> int:
        """resolution as the healpy-compatible nside parameter"""
        return 2**self.level

    @property
    def nest(self: Self) -> bool:
        """indexing_scheme as the healpy-compatible nest parameter"""
        if self.indexing_scheme not in {"nested", "ring"}:
            raise ValueError(
                f"cannot convert indexing scheme {self.indexing_scheme} to `nest`"
            )
        else:
            return self.indexing_scheme == "nested"

    @classmethod
    def from_dict(cls: type[T], mapping: dict[str, Any]) -> T:
        """construct a `HealpixInfo` object from a mapping of attributes

        Parameters
        ----------
        mapping: mapping of str to any
            The attributes.

        Returns
        -------
        grid_info : HealpixInfo
            The constructed grid info object.
        """

        def translate_nside(nside):
            log = np.log2(nside)
            potential_level = int(log)
            if potential_level != log:
                raise ValueError("`nside` has to be an integer power of 2")

            return potential_level

        translations = {
            "nside": ("level", translate_nside),
            "order": ("level", identity),
            "resolution": ("level", identity),
            "depth": ("level", identity),
            "nest": ("indexing_scheme", lambda nest: "nested" if nest else "ring"),
        }

        params = translate_parameters(mapping, translations)
        return cls(**params)

    def to_dict(self: Self) -> dict[str, Any]:
        """
        Dump the normalized grid parameters.

        Returns
        -------
        mapping : dict of str to any
            The normalized grid parameters.
        """
        return {
            "grid_name": "healpix",
            "level": self.level,
            "indexing_scheme": self.indexing_scheme,
        }

    def cell_ids2geographic(self, cell_ids):
        """
        Convert cell ids to geographic coordinates

        Parameters
        ----------
        cell_ids : array-like
            Array-like containing the cell ids.

        Returns
        -------
        lon : array-like
            The longitude coordinate values of the grid cells in degree
        lat : array-like
            The latitude coordinate values of the grid cells in degree
        """
        converters = {
            "nested": cdshealpix.nested.healpix_to_lonlat,
            "ring": lambda cell_ids, level: cdshealpix.ring.healpix_to_lonlat(
                cell_ids, nside=2**level
            ),
        }
        converter = converters[self.indexing_scheme]

        lon, lat = converter(cell_ids, self.level)

        return np.asarray(lon.to("degree")), np.asarray(lat.to("degree"))

    def geographic2cell_ids(self, lon, lat):
        """
        Convert cell ids to geographic coordinates

        This will perform a binning operation: any point within a grid cell will be assign
        that cell's ID.

        Parameters
        ----------
        lon : array-like
            The longitude coordinate values in degree
        lat : array-like
            The latitude coordinate values in degree

        Returns
        -------
        cell_ids : array-like
            Array-like containing the cell ids.
        """
        from astropy.coordinates import Latitude, Longitude

        converters = {
            "nested": cdshealpix.nested.lonlat_to_healpix,
            "ring": lambda lon, lat, level: cdshealpix.ring.lonlat_to_healpix(
                lon, lat, nside=2**level
            ),
        }
        converter = converters[self.indexing_scheme]

        longitude = Longitude(lon, unit="degree")
        latitude = Latitude(lat, unit="degree")

        return converter(longitude, latitude, self.level)

    def cell_boundaries(self, cell_ids: Any, backend="shapely") -> np.ndarray:
        """
        Derive cell boundary polygons from cell ids

        Parameters
        ----------
        cell_ids : array-like
            The cell ids.
        backend : {"shapely", "geoarrow"}, default: "shapely"
            The backend to convert to.

        Returns
        -------
        polygons : array-like
            The derived cell boundary polygons. The format differs based on the passed
            backend:

            - ``"shapely"``: return a array of :py:class:`shapely.Polygon` objects
            - ``"geoarrow"``: return a ``geoarrow`` array
        """
        converters = {
            "nested": cdshealpix.nested.vertices,
            "ring": lambda cell_ids, level, **kwargs: cdshealpix.ring.vertices(
                cell_ids, nside=2**level, **kwargs
            ),
        }
        converter = converters[self.indexing_scheme]

        lon_, lat_ = converter(cell_ids, self.level, step=1)

        lon = np.asarray(lon_.to("degree"))
        lat = np.asarray(lat_.to("degree"))

        lon_reshaped = np.reshape(lon, (-1, 4))
        lat_reshaped = np.reshape(lat, (-1, 4))

        lon_ = center_around_prime_meridian(lon_reshaped, lat_reshaped)

        vertices = np.stack((lon_, lat_reshaped), axis=-1)

        backends = {
            "shapely": polygons_shapely,
            "geoarrow": polygons_geoarrow,
        }

        backend_func = backends.get(backend)
        if backend_func is None:
            raise ValueError(f"invalid backend: {backend!r}")

        return backend_func(vertices)

    def zoom_to(self, cell_ids, level):
        if self.indexing_scheme == "ring":
            raise ValueError(
                "Scaling does not make sense for the 'ring' scheme."
                " Please convert to a nested scheme first."
            )

        from healpix_geo.nested import zoom_to

        return zoom_to(cell_ids, self.level, level)


@register_dggs("healpix")
class HealpixIndex(DGGSIndex):
    def __init__(
        self,
        cell_ids: Any | PandasIndex,
        dim: str,
        grid_info: DGGSInfo,
    ):
        if not isinstance(grid_info, HealpixInfo):
            raise ValueError(f"grid info object has an invalid type: {type(grid_info)}")

        super().__init__(cell_ids, dim, grid_info)

    @classmethod
    def from_variables(
        cls: type["HealpixIndex"],
        variables: Mapping[Any, xr.Variable],
        *,
        options: Mapping[str, Any],
    ) -> "HealpixIndex":
        _, var, dim = _extract_cell_id_variable(variables)

        grid_info = HealpixInfo.from_dict(var.attrs | options)

        return cls(var.data, dim, grid_info)

    def _replace(self, new_pd_index: PandasIndex):
        return type(self)(new_pd_index, self._dim, self._grid)

    @property
    def grid_info(self) -> HealpixInfo:
        return self._grid

    def _repr_inline_(self, max_width: int):
        return f"HealpixIndex(level={self._grid.level}, indexing_scheme={self._grid.indexing_scheme})"



import numpy.typing as npt
import xarray as xr

from xdggs.grid import DGGSInfo
from xdggs.index import DGGSIndex
from xdggs.plotting import explore


@xr.register_dataset_accessor("dggs")
@xr.register_dataarray_accessor("dggs")
class DGGSAccessor:
    _obj: xr.Dataset | xr.DataArray
    _index: DGGSIndex | None
    _name: str

    def __init__(self, obj: xr.Dataset | xr.DataArray):
        self._obj = obj

        index = None
        name = ""
        for k, idx in obj.xindexes.items():
            if isinstance(idx, DGGSIndex):
                if index is not None:
                    raise ValueError(
                        "Only one DGGSIndex per dataset or dataarray is supported"
                    )
                index = idx
                name = k
        self._name = name
        self._index = index

    def decode(self, grid_info=None, *, name="cell_ids") -> xr.Dataset | xr.DataArray:
        """decode the DGGS cell ids

        Parameters
        ----------
        grid_info : dict or DGGSInfo, optional
            Override the grid parameters on the dataset. Useful to set attributes on
            the dataset.
        name : str, default: "cell_ids"
            The name of the coordinate containing the cell ids.

        Returns
        -------
        obj : xarray.DataArray or xarray.Dataset
            The object with a DGGS index on the cell id coordinate.
        """
        var = self._obj[name]
        if isinstance(grid_info, DGGSInfo):
            grid_info = grid_info.to_dict()
        if isinstance(grid_info, dict):
            var.attrs = grid_info

        return self._obj.drop_indexes(name, errors="ignore").set_xindex(name, DGGSIndex)

    @property
    def index(self) -> DGGSIndex:
        """The DGGSIndex instance for this Dataset or DataArray.

        Raises
        ------
        ValueError
            if no DGGSIndex can be found
        """
        if self._index is None:
            raise ValueError("no DGGSIndex found on this Dataset or DataArray")
        return self._index

    @property
    def coord(self) -> xr.DataArray:
        """The indexed DGGS (cell ids) coordinate as a DataArray.

        Raises
        ------
        ValueError
            if no such coordinate is found on the Dataset / DataArray
        """
        if not self._name:
            raise ValueError(
                "no coordinate with a DGGSIndex found on this Dataset or DataArray"
            )
        return self._obj[self._name]

    @property
    def params(self) -> dict:
        """The grid parameters after normalization."""
        return self.index.grid.to_dict()

    @property
    def grid_info(self) -> DGGSInfo:
        """The grid info object containing the DGGS type and its parameters.

        Returns
        -------
        xdggs.DGGSInfo
        """
        return self.index.grid_info

    def sel_latlon(
        self, latitude: npt.ArrayLike, longitude: npt.ArrayLike
    ) -> xr.Dataset | xr.DataArray:
        """Select grid cells from latitude/longitude data.

        Parameters
        ----------
        latitude : array-like
            Latitude coordinates (degrees).
        longitude : array-like
            Longitude coordinates (degrees).

        Returns
        -------
        subset
            A new :py:class:`xarray.Dataset` or :py:class:`xarray.DataArray`
            with all cells that contain the input latitude/longitude data points.
        """
        cell_indexers = {
            self._name: self.grid_info.geographic2cell_ids(lon=longitude, lat=latitude)
        }
        return self._obj.sel(cell_indexers)

    def assign_latlon_coords(self) -> xr.Dataset | xr.DataArray:
        """Return a new Dataset or DataArray with new "latitude" and "longitude"
        coordinates representing the grid cell centers."""

        lon_data, lat_data = self.index.cell_centers()

        return self._obj.assign_coords(
            latitude=(self.index._dim, lat_data),
            longitude=(self.index._dim, lon_data),
        )

    @property
    def cell_ids(self):
        """The indexed DGGS (cell ids) coordinate as a DataArray.

        Alias of ``coord``.

        Raises
        ------
        ValueError
            if no such coordinate is found on the Dataset / DataArray
        """
        return self.coord

    def cell_centers(self):
        """derive geographic cell center coordinates

        Returns
        -------
        coords : xarray.Dataset
            Dataset containing the cell centers in geographic coordinates.
        """
        lon_data, lat_data = self.index.cell_centers()

        return xr.Dataset(
            coords={
                "latitude": (self.index._dim, lat_data),
                "longitude": (self.index._dim, lon_data),
            }
        )

    def cell_boundaries(self):
        """derive cell boundary polygons

        Returns
        -------
        boundaries : xarray.DataArray
            The cell boundaries as shapely objects.
        """
        boundaries = self.index.cell_boundaries()

        return xr.DataArray(
            boundaries, coords={self._name: self.cell_ids}, dims=self.cell_ids.dims
        )

    def zoom_to(self, level: int):
        """Change the refinement level of the cell ids to `level`.

        Parameters
        ----------
        level : int
            The refinement level to change to. Can be smaller than the dataset's
            level to compute parents, or bigger to fetch the children. In the
            latter case, the array will have an additional `"children"`
            dimension.

        Returns
        -------
        zoomed : xr.DataArray
            The children or parents of the current cells.
        """
        zoomed = self.index.zoom_to(level=level)

        if zoomed.ndim == 1:
            dims = self.cell_ids.dims
        else:
            dims = [*self.cell_ids.dims, "children"]

        return xr.DataArray(zoomed, coords={self._name: self.cell_ids}, dims=dims)

    def explore(self, *, cmap="viridis", center=None, alpha=None, coords=None):
        """interactively explore the data using `lonboard`

        Requires `lonboard`, `matplotlib`, and `arro3.core` to be installed.

        Parameters
        ----------
        cmap : str
            The name of the color map to use
        center : int or float, optional
            If set, will use this as the center value of a diverging color map.
        alpha : float, optional
            If set, controls the transparency of the polygons.
        coords : list of str, default: ["latitude", "longitude"]
            Additional coordinates to contain in the table of contents.

        Returns
        -------
        map : lonboard.Map
            The rendered map.

        Notes
        -----
        Plotting currently is restricted to 1D `DataArray` objects.
        """
        if isinstance(self._obj, xr.Dataset):
            raise ValueError("does not work with Dataset objects, yet")

        return explore(
            self._obj,
            cmap=cmap,
            center=center,
            alpha=alpha,
            coords=coords,
        )



import itertools

import numpy as np
import pytest
import shapely
import shapely.testing
import xarray as xr
from xarray.core.indexes import PandasIndex

from xdggs import h3
from xdggs.tests import geoarrow_to_shapely

# from the h3 gallery, at resolution 3
cell_ids = [
    np.array([0x832830FFFFFFFFF]),
    np.array([0x832831FFFFFFFFF, 0x832832FFFFFFFFF]),
    np.array([0x832833FFFFFFFFF, 0x832834FFFFFFFFF, 0x832835FFFFFFFFF]),
]
cell_centers = [
    np.array([[-122.19619676, 38.19320895]]),
    np.array([[-123.43390346, 38.63853196], [-121.00991811, 38.82387033]]),
    np.array(
        [
            [-122.2594399, 39.27846774],
            [-122.13425086, 37.09786649],
            [-123.35925909, 37.55231005],
        ]
    ),
]
dims = ["cells", "zones"]
levels = [1, 5, 15]
variable_names = ["cell_ids", "zonal_ids", "zone_ids"]

variables = [
    xr.Variable(dims[0], cell_ids[0], {"grid_name": "h3", "level": levels[0]}),
    xr.Variable(dims[1], cell_ids[0], {"grid_name": "h3", "level": levels[0]}),
    xr.Variable(dims[0], cell_ids[1], {"grid_name": "h3", "level": levels[1]}),
    xr.Variable(dims[1], cell_ids[2], {"grid_name": "h3", "level": levels[2]}),
]
variable_combinations = [
    (old, new) for old, new in itertools.product(variables, repeat=2)
]


class TestH3Info:

    @pytest.mark.parametrize(
        ["level", "error"],
        (
            (0, None),
            (1, None),
            (-1, ValueError("level must be an integer between")),
        ),
    )
    def test_init(self, level, error):
        if error is not None:
            with pytest.raises(type(error), match=str(error)):
                h3.H3Info(level=level)
            return

        actual = h3.H3Info(level=level)

        assert actual.level == level

    @pytest.mark.parametrize(
        ["mapping", "expected"],
        (
            ({"level": 0}, 0),
            ({"level": np.int64(2)}, 2),
            ({"resolution": 1}, 1),
            ({"level": -1}, ValueError("level must be an integer between")),
        ),
    )
    def test_from_dict(self, mapping, expected):
        if isinstance(expected, Exception):
            with pytest.raises(type(expected), match=str(expected)):
                h3.H3Info.from_dict(mapping)
            return

        actual = h3.H3Info.from_dict(mapping)
        assert actual.level == expected and type(actual.level) is type(expected)

    def test_roundtrip(self):
        mapping = {"grid_name": "h3", "level": 0}

        grid = h3.H3Info.from_dict(mapping)
        actual = grid.to_dict()

        assert actual == mapping

    @pytest.mark.parametrize(
        ["cell_ids", "cell_centers"], list(zip(cell_ids, cell_centers))
    )
    def test_cell_ids2geographic(self, cell_ids, cell_centers):
        grid = h3.H3Info(level=3)

        actual = grid.cell_ids2geographic(cell_ids)
        expected = cell_centers.T

        assert isinstance(actual, tuple) and len(actual) == 2
        np.testing.assert_allclose(actual, expected)

    @pytest.mark.parametrize(
        ["cell_centers", "cell_ids"], list(zip(cell_centers, cell_ids))
    )
    def test_geographic2cell_ids(self, cell_centers, cell_ids):
        grid = h3.H3Info(level=3)

        actual = grid.geographic2cell_ids(
            lon=cell_centers[:, 0], lat=cell_centers[:, 1]
        )
        expected = cell_ids

        np.testing.assert_equal(actual, expected)

    @pytest.mark.parametrize(
        ["level", "cell_ids", "expected_coords"],
        (
            (
                1,
                np.array([0x81283FFFFFFFFFF]),
                np.array(
                    [
                        [
                            [-121.70715692, 36.5742183],
                            [-119.00228227, 40.57057179],
                            [-122.13483148, 44.05769081],
                            [-127.95866237, 43.41920841],
                            [-130.22263777, 39.44242422],
                            [-127.13810062, 36.07979648],
                            [-121.70715692, 36.5742183],
                        ]
                    ]
                ),
            ),
            (
                3,
                np.array([0x832831FFFFFFFFF, 0x832832FFFFFFFFF]),
                np.array(
                    [
                        [
                            [-122.99764068, 38.13012709],
                            [-122.63105838, 38.7055711],
                            [-123.06903465, 39.21268153],
                            [-123.87318668, 39.14165748],
                            [-124.23124622, 38.566387],
                            [-123.7937797, 38.06195413],
                            [-122.99764068, 38.13012709],
                        ],
                        [
                            [-120.57845933, 38.30365554],
                            [-120.19218037, 38.87491264],
                            [-120.62501818, 39.3938676],
                            [-121.44467346, 39.33890002],
                            [-121.82297225, 38.76738776],
                            [-121.38971139, 38.25108747],
                            [-120.57845933, 38.30365554],
                        ],
                    ]
                ),
            ),
            (
                2,
                np.array([0x822837FFFFFFFFF, 0x821987FFFFFFFFF, 0x82285FFFFFFFFFF]),
                np.array(
                    [
                        [
                            [-121.70715692, 36.5742183],
                            [-120.15030816, 37.77836118],
                            [-120.62501818, 39.3938676],
                            [-122.69909887, 39.78423084],
                            [-124.23124622, 38.566387],
                            [-123.71598552, 36.97229615],
                            [-121.70715692, 36.5742183],
                        ],
                        [
                            [-21.86089163, 59.14600883],
                            [-24.48971137, 58.33329382],
                            [-24.24608918, 56.81195076],
                            [-21.53679367, 56.10124011],
                            [-18.98719147, 56.88329845],
                            [-19.06702945, 58.40493644],
                            [-21.86089163, 59.14600883],
                        ],
                        [
                            [-132.06227559, 43.79453729],
                            [-130.64994419, 45.0396523],
                            [-131.29015942, 46.41694831],
                            [-133.36750995, 46.52708205],
                            [-134.72528083, 45.27804582],
                            [-134.06255972, 43.92295857],
                            [-132.06227559, 43.79453729],
                        ],
                    ]
                ),
            ),
        ),
    )
    @pytest.mark.parametrize("backend", ["shapely", "geoarrow"])
    def test_cell_boundaries(self, level, cell_ids, backend, expected_coords):
        expected = shapely.polygons(expected_coords)

        grid = h3.H3Info(level=level)

        backends = {"shapely": lambda arr: arr, "geoarrow": geoarrow_to_shapely}
        converter = backends[backend]

        actual = grid.cell_boundaries(cell_ids, backend=backend)

        shapely.testing.assert_geometries_equal(converter(actual), expected)

    @pytest.mark.parametrize(
        ["level", "cell_ids", "new_level", "expected"],
        (
            pytest.param(
                3,
                np.array([0x832833FFFFFFFFF, 0x832834FFFFFFFFF, 0x832835FFFFFFFFF]),
                2,
                np.array([0x822837FFFFFFFFF, 0x822837FFFFFFFFF, 0x822837FFFFFFFFF]),
                id="parents",
            ),
            pytest.param(
                3,
                np.array([0x832833FFFFFFFFF, 0x832834FFFFFFFFF, 0x832835FFFFFFFFF]),
                1,
                np.array([0x81283FFFFFFFFFF, 0x81283FFFFFFFFFF, 0x81283FFFFFFFFFF]),
                id="grandparents",
            ),
        ),
    )
    def test_zoom_to(self, level, cell_ids, new_level, expected):
        grid = h3.H3Info(level=level)

        actual = grid.zoom_to(cell_ids, level=new_level)
        np.testing.assert_equal(actual, expected)

    def test_zoom_to_children_not_implemented(self):
        grid = h3.H3Info(level=3)
        cell_ids = np.arange(3)

        with pytest.raises(NotImplementedError):
            grid.zoom_to(cell_ids, level=4)


@pytest.mark.parametrize("level", levels)
@pytest.mark.parametrize("dim", dims)
@pytest.mark.parametrize("cell_ids", cell_ids)
def test_init(cell_ids, dim, level):
    grid = h3.H3Info(level)
    index = h3.H3Index(cell_ids, dim, grid)

    assert index._grid == grid
    assert index._dim == dim

    # TODO: how do we check the index, if at all?
    assert index._pd_index.dim == dim
    assert np.all(index._pd_index.index.values == cell_ids)


@pytest.mark.parametrize("level", levels)
def test_grid(level):
    grid = h3.H3Info(level)

    index = h3.H3Index([0], "cell_ids", grid)

    assert index.grid_info is grid


@pytest.mark.parametrize("variable", variables)
@pytest.mark.parametrize("variable_name", variable_names)
@pytest.mark.parametrize("options", [{}])
def test_from_variables(variable_name, variable, options):
    expected_level = variable.attrs["level"]

    variables = {variable_name: variable}
    index = h3.H3Index.from_variables(variables, options=options)

    assert index._grid.level == expected_level
    assert (index._dim,) == variable.dims

    # TODO: how do we check the index, if at all?
    assert (index._pd_index.dim,) == variable.dims
    assert np.all(index._pd_index.index.values == variable.data)


@pytest.mark.parametrize(["old_variable", "new_variable"], variable_combinations)
def test_replace(old_variable, new_variable):
    grid = h3.H3Info(level=old_variable.attrs["level"])
    index = h3.H3Index(
        cell_ids=old_variable.data,
        dim=old_variable.dims[0],
        grid_info=grid,
    )
    new_pandas_index = PandasIndex.from_variables(
        {"cell_ids": new_variable}, options={}
    )

    new_index = index._replace(new_pandas_index)

    assert new_index._grid == index._grid
    assert new_index._dim == index._dim
    assert new_index._pd_index == new_pandas_index


@pytest.mark.parametrize("max_width", [20, 50, 80, 120])
@pytest.mark.parametrize("level", levels)
def test_repr_inline(level, max_width):
    grid = h3.H3Info(level=level)
    index = h3.H3Index(cell_ids=[0], dim="cells", grid_info=grid)

    actual = index._repr_inline_(max_width)

    assert f"level={level}" in actual
    # ignore max_width for now
    # assert len(actual) <= max_width



import geoarrow.pyarrow as ga
import shapely

from xdggs.tests.matchers import (  # noqa: F401
    Match,
    MatchResult,
    assert_exceptions_equal,
)


def geoarrow_to_shapely(arr):
    return shapely.from_wkb(ga.as_wkb(arr))



import enum
import re
from dataclasses import dataclass, field


class MatchResult(enum.Enum):
    match = 0
    mismatched_typing = 1
    mismatched_message = 2


def is_exception_spec(exc):
    if not isinstance(exc, tuple):
        exc = (exc,)

    return all(isinstance(e, type) and issubclass(e, BaseException) for e in exc)


def is_exceptiongroup_spec(exc):
    if not isinstance(exc, tuple):
        exc = (exc,)

    return all(isinstance(e, type) and issubclass(e, BaseExceptionGroup) for e in exc)


class Match: ...


MatchType = BaseException | Match | tuple[BaseException | Match, ...]


def extract_message(exc):
    return getattr(exc, "message", str(exc))


# necessary until pytest-dev/pytest#11538 is resolved
@dataclass
class Match:
    """match exceptions and exception groups

    Think of `Match` objects as an equivalent to `re.Pattern` classes.

    Providing a tuple in place of a single exception / matcher means logical "or".

    Parameters
    ----------
    exc : exception-like or tuple of exception-like
        The exceptions or exception groups to match.
    submatchers : match-like, optional
        The submatchers for exception groups. Note that matchers with a mixture of
        exception groups and exceptions can't provide submatchers. If that's what you
        need, provide a tuple containing multiple matchers.
    match : str or regex-like, optional
        A pattern for matching the message of the exception.
    """

    exc: type[BaseException] | tuple[type[BaseException], ...]
    submatchers: list[MatchType] = field(default_factory=list)
    match: str = None

    def __post_init__(self):
        if not is_exception_spec(self.exc):
            raise TypeError(
                f"exception type must be one or more exceptions, got: {self.exc}"
            )
        if not is_exceptiongroup_spec(self.exc) and self.submatchers:
            raise TypeError("can only pass sub-matchers for exception groups")
        if not isinstance(self.match, str) and self.match is not None:
            raise TypeError("match must be either `None` or a string pattern")

    @classmethod
    def from_dict(cls, mapping):
        children = [cls.from_dict(m) for m in mapping.get("children", [])]

        return cls(
            mapping["exceptions"],
            submatchers=children,
            match=mapping.get("match", None),
        )

    def matches(self, exc):
        if not isinstance(exc, self.exc):
            return False

        if self.match is not None:
            message = extract_message(exc)
            match_ = re.search(self.match, message)
            if match_ is None:
                return False

        if self.submatchers and not isinstance(exc, BaseExceptionGroup):
            return False
        elif self.submatchers:
            if len(self.submatchers) != len(exc.exceptions):
                return False

            unmatched_matchers = []
            exceptions = list(exc.exceptions)
            for matcher in self.submatchers:
                for index, exception in enumerate(exceptions):
                    if matcher.matches(exception):
                        exceptions.pop(index)
                        break
                else:
                    unmatched_matchers.append(matcher)

            if unmatched_matchers:
                return False

        return True


def compare_exceptions(a, b):
    if type(a) is not type(b):
        return False

    if isinstance(a, ExceptionGroup):
        comparison = a.args[0] == b.args[0] and all(
            compare_exceptions(_a, _b) for _a, _b in zip(a.args[1], b.args[1])
        )
    else:
        comparison = a.args == b.args

    return comparison


def format_exception_diff(a, b):
    sections = []
    if type(a) is not type(b):
        sections.append("\n".join([f"L  {type(a).__name__}", f"R  {type(b).__name__}"]))
    else:
        sections.append(f"{type(a).__name__}")

    if isinstance(a, BaseExceptionGroup):
        if a.message != b.message:
            sections.append(
                "\n".join(["Message", f"L  {a.message}", f"R  {b.message}"])
            )
        if a.exceptions != b.exceptions:
            sections.append("\n".join(["Exceptions differ"]))
    elif str(a) != str(b):
        sections.append("\n".join(["Message", f"L  {str(a)}", f"R  {str(b)}"]))

    return "\n\n".join(sections)


def assert_exceptions_equal(actual, expected):
    assert compare_exceptions(actual, expected), format_exception_diff(actual, expected)



import json
from collections.abc import Mapping
from dataclasses import dataclass
from typing import Any, ClassVar, Self

import numpy as np
import xarray as xr

try:
    from h3ronpy import change_resolution
    from h3ronpy.vector import (
        cells_to_coordinates,
        cells_to_wkb_polygons,
        coordinates_to_cells,
    )
except ImportError:
    from h3ronpy.arrow import change_resolution
    from h3ronpy.arrow.vector import (
        cells_to_coordinates,
        cells_to_wkb_polygons,
        coordinates_to_cells,
    )
from xarray.indexes import PandasIndex

from xdggs.grid import DGGSInfo, translate_parameters
from xdggs.index import DGGSIndex
from xdggs.utils import _extract_cell_id_variable, register_dggs


def polygons_shapely(wkb):
    import shapely

    return shapely.from_wkb(wkb)


def polygons_geoarrow(wkb):
    import pyproj
    import shapely
    from arro3.core import list_array

    polygons = shapely.from_wkb(wkb)
    crs = pyproj.CRS.from_epsg(4326)

    geometry_type, coords, (ring_offsets, geom_offsets) = shapely.to_ragged_array(
        polygons
    )

    if geometry_type != shapely.GeometryType.POLYGON:
        raise ValueError(f"unsupported geometry type found: {geometry_type}")

    polygon_array = list_array(
        geom_offsets.astype("int32"), list_array(ring_offsets.astype("int32"), coords)
    )
    polygon_array_with_geo_meta = polygon_array.cast(
        polygon_array.field.with_metadata(
            {
                "ARROW:extension:name": "geoarrow.polygon",
                "ARROW:extension:metadata": json.dumps({"crs": crs.to_json_dict()}),
            }
        )
    )

    return polygon_array_with_geo_meta


@dataclass(frozen=True)
class H3Info(DGGSInfo):
    """
    Grid information container for h3 grids.

    Parameters
    ----------
    level : int
        Grid hierarchical level. A higher value corresponds to a finer grid resolution
        with smaller cell areas. The number of cells covering the whole sphere usually
        grows exponentially with increasing level values, ranging from 5-100 cells at
        level 0 to millions or billions of cells at level 10+ (the exact numbers depends
        on the specific grid).
    """

    level: int
    """int : The hierarchical level of the grid"""

    valid_parameters: ClassVar[dict[str, Any]] = {"level": range(16)}

    def __post_init__(self):
        if self.level not in self.valid_parameters["level"]:
            raise ValueError("level must be an integer between 0 and 15")

    @classmethod
    def from_dict(cls: type[Self], mapping: dict[str, Any]) -> Self:
        """construct a `H3Info` object from a mapping of attributes

        Parameters
        ----------
        mapping: mapping of str to any
            The attributes.

        Returns
        -------
        grid_info : H3Info
            The constructed grid info object.
        """
        translations = {
            "resolution": ("level", int),
            "level": ("level", int),
        }

        params = translate_parameters(mapping, translations)
        return cls(**params)

    def to_dict(self: Self) -> dict[str, Any]:
        """
        Dump the normalized grid parameters.

        Returns
        -------
        mapping : dict of str to any
            The normalized grid parameters.
        """
        return {"grid_name": "h3", "level": self.level}

    def cell_ids2geographic(
        self, cell_ids: np.ndarray
    ) -> tuple[np.ndarray, np.ndarray]:
        """
        Convert cell ids to geographic coordinates

        Parameters
        ----------
        cell_ids : array-like
            Array-like containing the cell ids.

        Returns
        -------
        lon : array-like
            The longitude coordinate values of the grid cells in degree
        lat : array-like
            The latitude coordinate values of the grid cells in degree
        """
        result = cells_to_coordinates(cell_ids, radians=False)

        lon = result.column("lng").to_numpy()
        lat = result.column("lat").to_numpy()

        return lon, lat

    def geographic2cell_ids(self, lon, lat):
        """
        Convert cell ids to geographic coordinates

        This will perform a binning operation: any point within a grid cell will be assign
        that cell's ID.

        Parameters
        ----------
        lon : array-like
            The longitude coordinate values in degree
        lat : array-like
            The latitude coordinate values in degree

        Returns
        -------
        cell_ids : array-like
            Array-like containing the cell ids.
        """
        return coordinates_to_cells(lat, lon, self.level, radians=False)

    def cell_boundaries(self, cell_ids, backend="shapely"):
        """
        Derive cell boundary polygons from cell ids

        Parameters
        ----------
        cell_ids : array-like
            The cell ids.
        backend : {"shapely", "geoarrow"}, default: "shapely"
            The backend to convert to.

        Returns
        -------
        polygons : array-like
            The derived cell boundary polygons. The format differs based on the passed
            backend:

            - ``"shapely"``: return a array of :py:class:`shapely.Polygon` objects
            - ``"geoarrow"``: return a ``geoarrow`` array
        """
        # TODO: convert cell ids directly to geoarrow once h3ronpy supports it
        wkb = cells_to_wkb_polygons(cell_ids, radians=False, link_cells=False)

        backends = {
            "shapely": polygons_shapely,
            "geoarrow": polygons_geoarrow,
        }
        backend_func = backends.get(backend)
        if backend_func is None:
            raise ValueError(f"invalid backend: {backend!r}")
        return backend_func(wkb)

    def zoom_to(self, cell_ids, level):
        if level > self.level:
            raise NotImplementedError(
                "extracting children is not supported for H3, yet."
            )

        return np.asarray(change_resolution(cell_ids, level))


@register_dggs("h3")
class H3Index(DGGSIndex):
    _grid: DGGSInfo

    def __init__(
        self,
        cell_ids: Any | PandasIndex,
        dim: str,
        grid_info: DGGSInfo,
    ):
        super().__init__(cell_ids, dim, grid_info)

    @classmethod
    def from_variables(
        cls: type["H3Index"],
        variables: Mapping[Any, xr.Variable],
        *,
        options: Mapping[str, Any],
    ) -> "H3Index":
        _, var, dim = _extract_cell_id_variable(variables)

        grid_info = H3Info.from_dict(var.attrs | options)

        return cls(var.data, dim, grid_info)

    @property
    def grid_info(self) -> H3Info:
        return self._grid

    def _replace(self, new_pd_index: PandasIndex):
        return type(self)(new_pd_index, self._dim, self._grid)

    def _repr_inline_(self, max_width: int):
        return f"H3Index(level={self._grid.level})"
